import java.awt.Image;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Event;
import java.awt.Point;

import java.applet.Applet;

import java.util.Random;
import java.util.Arrays;

// TODO:
//   - should really show lines between rows instead of using a row as a wall

public class Maze extends Applet implements Runnable{
	private int Width  = 0;
	private int Height = 0;
	private int MAZE_WIDTH  = Width / 5;
	private int MAZE_HEIGHT = Height / 5;

	private Point pStart = new Point(1, 1);
	private Point pFinish = new Point();

	private Point player = new Point(pStart);

	private int maze[][] = null;

	// keys
	private final int NONE  = -1;
	private final int DOWN  =  0;
	private final int UP    =  1;
	private final int LEFT  =  2;
	private final int RIGHT =  3;
	private final int SOLVE =  4;
	private final int NEW   =  5;
	private final int RESET =  6;

	private final int NEW_ADDWALL =  7;
	private final int NEW_REMWALL =  8;
	private final int NEW_TUNNEL  =  9;

	private Thread thread;

	private int keyPressed = NONE;
	private int mazeType   = NEW_ADDWALL;

	// double buffering
	private Image offscreenImage;
	private Graphics offscreenGraphics;

	private boolean move(Point p, int direction){
		boolean moved = false;

		switch(direction){
			case LEFT:
				if(maze[p.x - 1][p.y] == 0){
					p.x--;
					moved = true;
				}
				break;
			case RIGHT:
				if(maze[p.x + 1][p.y] == 0){
					p.x++;
					moved = true;
				}
				break;
			case UP:
				if(maze[p.x][p.y - 1] == 0){
					p.y--;
					moved = true;
				}
				break;
			case DOWN:
				if(maze[p.x][p.y + 1] == 0){
					p.y++;
					moved = true;
				}
				break;
			default:
				break;
		}

		return moved;
	}

	private boolean testDirection(Point p, int direction){
		boolean goodDir = false;

		switch(direction){
			case LEFT:
				if(maze[p.x - 1][p.y] == 0)
					goodDir = true;
				break;
			case RIGHT:
				if(maze[p.x + 1][p.y] == 0)
					goodDir = true;
				break;
			case UP:
				if(maze[p.x][p.y - 1] == 0)
					goodDir = true;
				break;
			case DOWN:
				if(maze[p.x][p.y + 1] == 0)
					goodDir = true;
				break;
			default:
				break;
		}

		return goodDir;
	}

	private int getBestDirection(Point mouse, boolean firstPass, int dir){
		int tries[] = null;
		if(firstPass){
			if(dir == NONE)
				dir = RIGHT;
			// follow left wall
			tries = new int[] {UP, RIGHT, DOWN, LEFT};
		}
		else{
			if(dir == NONE)
				dir = LEFT;
			// follow right wall
			tries = new int[] {UP, LEFT, DOWN, RIGHT};
		}

		int index = 0;
		for(; index < 4; index++){
			if(dir == tries[index])
				break;
		}
		if(--index < 0)
			index = 3;

		dir = NONE;
		for(int i = 0; i < 4; i++){
			int j = (index + i) % 4;
			if(testDirection(mouse, tries[j])){
				dir = tries[j];
				break;
			}
		}

		return dir;
	}

	private boolean solveMaze(boolean watch){
		Point mouse = new Point(player);
		boolean goodMaze = true;
		boolean firstPass = true;
		int dir = NONE;

		while(!mouse.equals(pFinish) && goodMaze){
			dir = getBestDirection(mouse, firstPass, dir);

			if(dir == NONE){
				if(firstPass){
					firstPass = false;
					mouse.setLocation(player);
					continue;
				}
				else
					return false;
			}

			move(mouse, dir);

			if(mouse.equals(player)){
				if(firstPass)
					firstPass = false;
				else
					goodMaze = false;
			}

			//watch the computer solve the maze?
			if(watch){
				render(mouse, Color.red);

				try{
					Thread.sleep(100);
				}
				catch(Exception e){
					showStatus("Sleep problem");
				}
			}
		}

		// set player to finish location
//		if(watch)
//			player.setLocation(mouse);

		return goodMaze;
	}

	private void addRandomWall(){
		Random random = new Random(System.currentTimeMillis());

		// random wall creation
		for(int j = 1; j < (MAZE_HEIGHT - 1); j++){
			for(int i = 1; i < (MAZE_WIDTH - 1); i++){
				if((i != pStart.x || j != pStart.y) && random.nextFloat() > 0.5){
					maze[i][j] = 1;
					if(!solveMaze(false))
						maze[i][j] = 0;
				}
			}
		}
	}

	private void removeRandomWall(){
		Random random = new Random(System.currentTimeMillis());

		for(int i = 0; i < MAZE_WIDTH; i++)
			Arrays.fill(maze[i], 1);
		maze[pStart.x][pStart.y] = maze[pFinish.x][pFinish.y] = 0;

		// random wall removal
		while(!solveMaze(false)){
			int x = random.nextInt(MAZE_WIDTH - 2) + 1;
			int y = random.nextInt(MAZE_HEIGHT - 2) + 1;
			maze[x][y] = 0;
		}
	}

	private void tunnelFrom(Point p){
		Random random = new Random(System.currentTimeMillis());
		Point mouse = new Point(p);

		for(int i = 0; i < MAZE_WIDTH; i++)
			Arrays.fill(maze[i], 1);
		maze[pStart.x][pStart.y] = maze[pFinish.x][pFinish.y] = 0;

		while(!mouse.equals(pFinish)){
			maze[mouse.x][mouse.y] = 0;
			float rnd = random.nextFloat();
			if(rnd < 0.5){
				int delta = pFinish.x - mouse.x;
				mouse.x += delta > 0 ? 1 : (delta < 0 ? -1 : 0);
			}
			else{
				int delta = pFinish.y - mouse.y;
				mouse.y += delta > 0 ? 1 : (delta < 0 ? -1 : 0);
			}
		}

		// remove a random number of points for effect
		int maxPoints = MAZE_HEIGHT * MAZE_WIDTH / 2;
		int numPoints = random.nextInt(maxPoints);
		for(int i = 0; i < numPoints; i++)
			maze[random.nextInt(MAZE_WIDTH - 2) + 1][random.nextInt(MAZE_HEIGHT - 2) + 1] = 0;
	}

	private void generateMaze(){
		pStart = new Point(1, 1);
		pFinish = new Point(MAZE_WIDTH - 2, MAZE_HEIGHT - 2);

		player = new Point(pStart);

		maze = new int[MAZE_WIDTH][MAZE_HEIGHT];

		//create bounding box
		for(int i = 0; i < MAZE_WIDTH; i++)
			maze[i][0] = maze[i][MAZE_HEIGHT - 1] = 1;
		for(int i = 0; i < MAZE_HEIGHT; i++)
			maze[0][i] = maze[MAZE_WIDTH - 1][i] = 1;

		switch(mazeType){
			case NEW_REMWALL:
				removeRandomWall();
				break;
			case NEW_TUNNEL:
				tunnelFrom(pStart);
//				tunnelFrom(pFinish);
				break;
			case NEW_ADDWALL:
			default:
				addRandomWall();
				break;
		}
	}

	public void init(){
		Width  = getWidth();
		Height = getHeight();
		MAZE_WIDTH  = Width / 5;
		MAZE_HEIGHT = Height / 5;
	}

	public void start(){
		thread = new Thread(this);
		thread.start();
	}

	public void run(){
		requestFocus();

		// create offscreen buffer
		offscreenImage = createImage(Width, Height);
		offscreenGraphics = offscreenImage.getGraphics();

		generateMaze();

		while(true){
			switch(keyPressed){
				case LEFT:
				case RIGHT:
				case UP:
				case DOWN:
					move(player, keyPressed);
					break;
				case SOLVE:
					solveMaze(true);
					break;
				case NEW:
					generateMaze();
					break;
				case NEW_ADDWALL:
					mazeType = NEW_ADDWALL;
					generateMaze();
					break;
				case NEW_REMWALL:
					mazeType = NEW_REMWALL;
					generateMaze();
					break;
				case NEW_TUNNEL:
					mazeType = NEW_TUNNEL;
					generateMaze();
					break;
				case RESET:
					player.setLocation(pStart);
					break;
				case NONE:
				default:
					break;
			}

			render(player, Color.red);
			try{
				Thread.sleep(50);
			}
			catch(Exception sleepProblem){
				showStatus("Sleep problem");
			}
		}
	}

	private void drawBackground(){
		int c = 25;
		int delta = 10;
		int shadeHeight = Height * delta / (2 * 255);
		for(int i = 0; i < Height; i += shadeHeight){
			offscreenGraphics.setColor(new Color(c, c, 255));
			offscreenGraphics.fillRect(0, i, Width, shadeHeight);
			c += delta;
			if(c > 255){
				c = 255;
				delta = -10;
			}
			else if(c < 0){
				c = 0;
				delta = 10;
			}
		}
	}

	private void drawMaze(){
		offscreenGraphics.setColor(Color.black);
		for(int i = 0; i < MAZE_WIDTH; i++){
			offscreenGraphics.fillRect(i*5,0,5,5);
			offscreenGraphics.fillRect(i*5,(MAZE_HEIGHT-1)*5,5,5);
		}
		for(int i = 1; i < MAZE_HEIGHT; i++){
			offscreenGraphics.fillRect(0,i*5,5,5);
			offscreenGraphics.fillRect((MAZE_WIDTH-1)*5,i*5,5,5);
		}
		for(int i = 1; i < (MAZE_WIDTH - 1); i++){
			for(int j = 1; j < (MAZE_HEIGHT - 1); j++){
				if(maze[i][j] == 1){
/*
					if(maze[i-1][j] == 1)
						offscreenGraphics.drawLine(i*5,(j+1)*5,(i+1)*5,(j+1)*5);
					if(maze[i+1][j] == 1)
						offscreenGraphics.drawLine((i+1)*5,(j+1)*5,(i+2)*5,(j+1)*5);
					if(maze[i][j-1] == 1)
						offscreenGraphics.drawLine((i+1)*5,j*5,(i+1)*5,(j+1)*5);
					if(maze[i][j+1] == 1)
						offscreenGraphics.drawLine((i+1)*5,(j+1)*5,(i+1)*5,(j+2)*5);
*/
					offscreenGraphics.fillRect(i*5,j*5,5,5);
				}
			}
		}

		offscreenGraphics.setColor(Color.yellow);
		offscreenGraphics.fillRect(pFinish.x * 5, pFinish.y * 5, 5, 5);
	}

	private void render(Point p, Color c){
		drawBackground();
		drawMaze();

		offscreenGraphics.setColor(c);
		offscreenGraphics.fillRect(p.x * 5, p.y * 5, 5, 5);

		// blit to screen
		paint(getGraphics());
	}

	public void paint(Graphics g){
		if(offscreenImage != null)
			g.drawImage(offscreenImage, 0, 0, this);
	}

	public void stop(){
		if((thread != null) && thread.isAlive())
			thread = null;
	}

	public boolean keyDown(Event evt, int key){
		switch(key){
			case Event.UP:
			case 'i':
			case 'I':
				keyPressed = UP;
				break;
			case Event.DOWN:
			case 'k':
			case 'K':
				keyPressed = DOWN;
				break;
			case Event.LEFT:
			case 'j':
			case 'J':
				keyPressed = LEFT;
				break;
			case Event.RIGHT:
			case 'l':
			case 'L':
				keyPressed = RIGHT;
				break;
			case 's':
			case 'S':
				keyPressed = SOLVE;
				break;
			case 'n':
			case 'N':
				keyPressed = NEW;
				break;
			case 'r':
			case 'R':
				keyPressed = RESET;
				break;
			case '1':
				keyPressed = NEW_ADDWALL;
				break;
			case '2':
				keyPressed = NEW_REMWALL;
				break;
			case '3':
				keyPressed = NEW_TUNNEL;
				break;
			default:
				break;
		}

		return true;
	}

	public boolean keyUp(Event evt, int key){
		switch(key){
			case Event.UP:
			case 'i':
			case 'I':
			case Event.DOWN:
			case 'k':
			case 'K':
			case Event.LEFT:
			case 'j':
			case 'J':
			case Event.RIGHT:
			case 'l':
			case 'L':
			case 's':
			case 'S':
			case 'n':
			case 'N':
			case 'r':
			case 'R':
			case '1':
			case '2':
			case '3':
				keyPressed = NONE;
				break;
			default:
		}

		return true;
	}
}

// vim:ts=3:noet
